#!/usr/bin/env python

import RPi.GPIO as GPIO
import sys
import struct
import time

INSTR_IDCODE  = 0x01
INSTR_IMPCODE = 0x03
INSTR_ADDRESS = 0x08
INSTR_BYPASS  = 0xff
INSTR_DATA    = 0x09
INSTR_CONTROL = 0x0A
INSTR_SAMPLE  = 0xfffffff8

TOF           = (1 << 1 )
TIF           = (1 << 2 )
BRKST         = (1 << 3 )
DLOCK         = (1 << 5 )
DRWN          = (1 << 9 )
DERR          = (1 << 10)
DSTRT         = (1 << 11)
JTAGBRK       = (1 << 12)
SETDEV        = (1 << 14)
PROBTRAP      = (1 << 14)
PROBEN        = (1 << 15)
PRRST         = (1 << 16)
DMAACC        = (1 << 17)
PRACC         = (1 << 18)
PRNW          = (1 << 19)
PERRST        = (1 << 20)
HALT          = (1 << 21)
DOZE          = (1 << 22)
SYNC          = (1 << 23)
DNM           = (1 << 28)
ROCC          = (1 << 31)
DMA_WORD      = 0x00000100
MIPS_HALFWORD = 0x00000080
DMA_BYTE      = 0x00000000

BRCM_SPI_READ         = 0x1fc00000
BRCM_SPI_MMR          = 0x00000000
BRCM_SPI_MMR_SIZE     = 0
BRCM_SPI_CTL          = 0x18000040
BRCM_FLASHADDRESS     = 0x18000044
BRCM_SPI_DATA         = 0x18000048
BRCM_SPI_OPCODE       = 0x18000044
BRCM_SPI_CTL_START    = 0x80000000
BRCM_SPI_CTL_BUSY     = 0x80000000

spi_flash_read = BRCM_SPI_READ
spi_flash_mmr = BRCM_SPI_MMR
spi_flash_mmr_size = BRCM_SPI_MMR_SIZE
spi_flash_ctl = BRCM_SPI_CTL
spi_flash_opcode = BRCM_SPI_OPCODE
spi_flash_data = BRCM_SPI_DATA
spi_ctl_start = BRCM_SPI_CTL_START
spi_ctl_busy = BRCM_SPI_CTL_BUSY

SPI_WRITE_ENABLE      = 0
SPI_WRITE_DISABLE     = 1
SPI_RD_STATUS         = 2
SPI_WR_STATUS         = 3
SPI_RD_DATA           = 4
SPI_FAST_RD_DATA      = 5
SPI_PAGE_PROGRAM      = 6
SPI_SECTOR_ERASE      = 7
SPI_BULK_ERASE        = 8
SPI_DEEP_PWRDOWN      = 9
SPI_RD_SIG            = 10
SPI_RD_ID             = 11
BCM_SPI_WRITE_ENABLE  = 12
BCM_SPI_RD_STATUS     = 13
BCM_SPI_PAGE_PROGRAM  = 14
BCM_SPI_SECTOR_ERASE  = 15
BCM_SPI_RD_ID         = 16
SPI_MAX_OPCODES       = 17

STM_OP_WR_ENABLE         = 0x06   
STM_OP_WR_DISABLE        = 0x04   
STM_OP_RD_STATUS         = 0x05   
STM_OP_WR_STATUS         = 0x01   
STM_OP_RD_DATA           = 0x03   
STM_OP_FAST_RD_DATA      = 0x0b   
STM_OP_PAGE_PGRM         = 0x02   
STM_OP_SECTOR_ERASE      = 0xd8   
STM_OP_BULK_ERASE        = 0xc7   
STM_OP_DEEP_PWRDOWN      = 0xb9   
STM_OP_RD_SIG            = 0xab   
STM_OP_RD_ID             = 0x9f   
BCM_STM_OP_WR_ENABLE     = 0x0006
BCM_STM_OP_RD_STATUS     = 0x0105
BCM_STM_OP_PAGE_PGRM     = 0x0402
BCM_STM_OP_SECTOR_ERASE  = 0x02d8
BCM_STM_OP_RD_ID         = 0x049f

stm_opcodes = \
[
    [STM_OP_WR_ENABLE,              1, 0],
    [STM_OP_WR_DISABLE,             1, 0],
    [STM_OP_RD_STATUS,              1, 1],
    [STM_OP_WR_STATUS,              1, 0],
    [STM_OP_RD_DATA,                4, 4],
    [STM_OP_FAST_RD_DATA,           1, 0],
    [STM_OP_PAGE_PGRM,              8, 0],
    [STM_OP_SECTOR_ERASE,           4, 0],
    [STM_OP_BULK_ERASE,             1, 0],
    [STM_OP_DEEP_PWRDOWN,           1, 0],
    [STM_OP_RD_SIG,                 4, 1],
    [STM_OP_RD_ID,                  1, 3],
    [BCM_STM_OP_WR_ENABLE,          1, 0],
    [BCM_STM_OP_RD_STATUS,          1, 1],
    [BCM_STM_OP_PAGE_PGRM,          8, 0],
    [BCM_STM_OP_SECTOR_ERASE,       4, 0],
    [BCM_STM_OP_RD_ID,              1, 3]
]

def get_bits(data):
    length = len(hex(data)) - 2
    length = length // 2 + length % 2
    return length * 8

class JTAG:

    TDI = 6
    TDO = 13
    TMS = 19
    TCK = 26
    CURINSTR = None
    INSTR_LENGTH = 8

    def __init__(self):
        self.__init_gpio()
        self.reset()

    def clock(self, tms, tdi):
        tms = 1 if tms else 0
        tdi = 1 if tdi else 0
        GPIO.output(self.TCK, 0)
        GPIO.output(self.TMS, tms)
        GPIO.output(self.TDI, tdi)
        GPIO.output(self.TCK, 1)
        return GPIO.input(self.TDO)

    def WriteData(self, data):
        self.ReadWriteData(data)

    def ReadData(self):
        return self.ReadWriteData(0)

    def ReadWriteData(self, data):
        out_data = 0
        self.clock(1, 0)
        self.clock(0, 0)
        self.clock(0, 0)
        for i in range(32):            
            out_bit  = self.clock((i == 31), ((data >> i) & 1))
            out_data = out_data | (out_bit << i)
        self.clock(1,0)
        self.clock(0,0)
        return out_data

    def reset(self):
        self.clock(1, 0)
        self.clock(1, 0)
        self.clock(1, 0)
        self.clock(1, 0)
        self.clock(1, 0)
        self.clock(0, 0)

    def set_instr(self, instr):
        if instr == self.CURINSTR:
            return
        self.clock(1, 0)  
        self.clock(1, 0)
        self.clock(0, 0)
        self.clock(0, 0)
        for i in range(self.INSTR_LENGTH):
            self.clock(i==(self.INSTR_LENGTH - 1), (instr>>i)&1)
        self.clock(1, 0)
        self.clock(0, 0)

        self.CURINSTR = instr

    def __init_gpio(self):
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup([self.TDI, self.TMS, self.TCK], GPIO.OUT)
        GPIO.setup(self.TDO, GPIO.IN)

    def imcode(self):
        self.set_instr(INSTR_IMPCODE)
        print(hex(self.ReadData()))

    def address(self):
        self.set_instr(INSTR_ADDRESS)
        print(hex(self.ReadData()))

    def data_instr(self):
        self.set_instr(INSTR_DATA)
        print(hex(self.ReadData()))

    def idcode(self):
        self.set_instr(INSTR_IDCODE)
        print(INSTR_IDCODE, hex(self.ReadData()))

    def bypass(self, data):
        length = get_bits(data)
        out_data = 0
        self.set_instr(INSTR_BYPASS)
        self.clock(1, 0)
        self.clock(0, 0)
        self.clock(0, 0)
        for i in range(length + 1):            
            out_bit  = self.clock((i == length), ((data >> i) & 1))
            # print("in: %d, out: %d"%((data >> i) & 1 ,out_bit))
            out_data = out_data | (out_bit << i)
        out_data >>= 1
        self.clock(1,0)
        self.clock(0,0)
        print(hex(out_data))
        return out_data

    def run(self):
        self.set_instr(INSTR_BYPASS)
        print(hex(self.ReadWriteData(0x01010101)))
        print(hex(self.ReadData()))

    def read_flash_test(self, addr, mode):
        self.set_instr(INSTR_ADDRESS)
        self.WriteData(addr)
        self.set_instr(INSTR_CONTROL)
        ctrl_reg = self.ReadWriteData(DMAACC | DRWN | mode | DSTRT | PROBEN | PRACC)
        # print("ctrl_reg:", hex(ctrl_reg))
        self.set_instr(INSTR_DATA)
        data = self.ReadData()
        self.set_instr(INSTR_CONTROL)
        assert self.ReadWriteData(PROBEN | PRACC) & DERR == 0
        return data

    def write_cmd(self, addr, data, mode):
        # data = 0x41424344
        self.set_instr(INSTR_ADDRESS)
        self.WriteData(addr)
        self.set_instr(INSTR_DATA)
        self.WriteData(data)
        self.set_instr(INSTR_CONTROL)
        self.ReadWriteData(DMAACC | mode | DSTRT | PROBEN | PRACC)
        ctrl_reg = self.ReadWriteData(DMAACC | PROBEN | PRACC) & DSTRT
        while ctrl_reg:
            pass
        # print("ctrl_reg:", hex(ctrl_reg))
        self.set_instr(INSTR_CONTROL)
        if self.ReadWriteData(PROBEN | PRACC) & DERR:
            print("DMA Write Addr = %08x  Data = ERROR ON WRITE"%addr)
            raw_input()

    def sample(self):
        self.set_instr(INSTR_SAMPLE)
        print(hex(self.ReadWriteData(0x10101010)))

    def debug(self):
        self.set_instr(INSTR_CONTROL)
        self.ReadWriteData(PRACC | PROBEN | SETDEV | JTAGBRK )
        print(self.ReadWriteData(PRACC | PROBEN | SETDEV) & BRKST)
        
    # def spiflash_sendcmd(self, cmd):
    #     ptr_opcode = stm_opcodes[cmd]
    #     self.write_cmd(0x18000040, 0x0000, DMA_WORD)
    #     reg = self.read_flash_test(spi_flash_ctl, DMA_WORD)
    #     # print("reg & busy:", hex(reg & spi_ctl_busy))
    #     self.write_cmd(spi_flash_opcode, ptr_opcode[0], DMA_WORD)
    #     reg = (reg  & 0xffffff00) | ptr_opcode[0] | spi_ctl_start
    #     self.write_cmd(spi_flash_ctl, reg, DMA_WORD)
    #     reg = self.read_flash_test(spi_flash_ctl, DMA_WORD)
    #     # print("reg & busy:", hex(reg & spi_ctl_busy))
    #     if ptr_opcode[2] > 0:
    #         reg = self.read_flash_test(spi_flash_data, DMA_WORD)
    #         if ptr_opcode[2] == 1:
    #             mask = 0x000000ff
    #         elif ptr_opcode[2] == 2:
    #             mask = 0x0000ffff
    #         elif ptr_opcode[2] == 3:
    #             mask = 0x00ffffff
    #         else:
    #             mask = 0xffffffff
    #         reg &= mask
    #     else:
    #         reg = 0
    #     print("reg:", hex(reg))
    #     return reg

    def get_flashID(self):
        self.write_cmd(0x1fc00000 + 0xAAA, 0xaaAAaaAA, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + 0x554, 0x55555555, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + 0xAAA, 0x90909090, MIPS_HALFWORD)
        data = self.read_flash_test(0x1FC00000, DMA_WORD)
        manufacturer_code = data & 0xffff
        device_code = (data >> 16) & 0xffff
        print("manufacture code: ", hex(manufacturer_code))
        print("device_code: ", hex(device_code))
        self.write_cmd(0x1fc00000, 0x00F000F0, MIPS_HALFWORD)

    # def write_cmd2(self, addr, data):
    #     self.write_cmd(spi_flash_ctl, 0x000, DMA_WORD)
    #     self.spiflash_sendcmd(BCM_SPI_WRITE_ENABLE)
    #     reg = self.read_flash_test(spi_flash_ctl, DMA_WORD)
    #     print("reg & busy:", hex(reg & spi_ctl_busy))
    #     self.write_cmd(spi_flash_data, data, DMA_WORD)
    #     self.write_cmd(spi_flash_opcode, addr, DMA_WORD)
    #     reg = (reg & 0xffffff00) | 0x0402 | spi_ctl_start
    #     self.write_cmd(spi_flash_ctl, reg, DMA_WORD)
    #     reg = self.read_flash_test(spi_flash_ctl, DMA_WORD)
    #     print("reg & busy:", hex(reg & spi_ctl_busy))
    #     reg = self.spiflash_sendcmd(BCM_SPI_RD_STATUS)
    #     print("reg & SPI_STATUS_WIP:", reg & 0x01)

    def erase_flash(self, addr):
        self.write_cmd(0x1fc00000, 0x00F000F0, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00AA00AA, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x2AA << 1), 0x00550055, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00800080, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00AA00AA, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x2AA << 1), 0x00550055, MIPS_HALFWORD)    
        self.write_cmd(addr, 0x00300030, MIPS_HALFWORD)
        while self.read_flash_test(addr, MIPS_HALFWORD) & 0x80 != 0x80:
            pass

    def flash_poll(self, addr, data):
        while self.read_flash_test(addr, MIPS_HALFWORD) & 0x80 != data & 0x80:
            pass

    def write_flash(self, addr, data):
        if addr % 0x4000 == 0:
            # earse
            self.erase_flash(addr)
        self.write_flash_no_earse(addr, data)

    def write_flash_no_earse(self, addr, data):
        data_lo = (data & 0xFFFF)
        data_hi = ((data >> 16) & 0xFFFF)
        # reset
        self.write_cmd(0x1fc00000, 0x00F000F0, MIPS_HALFWORD)
        # write half word
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00AA00AA, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x2AA << 1), 0x00550055, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00A000A0, MIPS_HALFWORD)    
        self.write_cmd(addr + 2, data, MIPS_HALFWORD)
        self.flash_poll(addr, data_lo)

        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00AA00AA, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x2AA << 1), 0x00550055, MIPS_HALFWORD)    
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00A000A0, MIPS_HALFWORD)    
        self.write_cmd(addr, data, MIPS_HALFWORD)
        self.flash_poll(addr + 2, data_hi)

    def write_flash_bypass(self, addr, data):
        if addr % 0x4000 == 0:
            self.erase_flash(addr)
        self.write_flash_bypass_no_earse(addr, data)

    def write_flash_bypass_no_earse(self, addr, data):
        data_lo = (data & 0xFFFF)
        data_hi = ((data >> 16) & 0xFFFF)
        # reset
        self.write_cmd(0x1fc00000, 0x00F000F0, MIPS_HALFWORD)
        # write
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00A000A0, MIPS_HALFWORD)
        self.write_cmd(addr + 2, data, MIPS_HALFWORD)
        time.sleep(0.0000001)
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00A000A0, MIPS_HALFWORD)
        self.write_cmd(addr, data, MIPS_HALFWORD)
        time.sleep(0.0000001)

    def write_cfe(self):
        f = open("CFE.BIN", "rb")
        start_addr = 0x1fc00000
        while True:
            data = f.read(4)
            if data == "":
                break
            data = struct.unpack(">L", data)[0]
            self.write_flash_no_earse(start_addr, data)
            # print("0x%x Done"%start_addr)
            # print(hex(self.read_flash_test(start_addr, DMA_WORD)))
            start_addr += 4

    def unlock_bypass(self):
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00900090, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00aa00aa, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + (0x2aa << 1), 0x00550055, MIPS_HALFWORD)
        self.write_cmd(0x1fc00000 + (0x555 << 1), 0x00200020, MIPS_HALFWORD)
            
def str_to_int(data):
    if data.startswith("0x"):
        data = int(data, 16)
    else:
        data = int(data)
    return data

def main():
    # global INSTR_IDCODE
    # for x in xrange(0xff):
    #     INSTR_IDCODE = x
    tj = JTAG()
    tj.get_flashID()
    tj.idcode()
    tj.bypass(0x44434241)
    if len(sys.argv) < 2:
        return
    cmd = sys.argv[1]
    if cmd == "read":
        addr = str_to_int(sys.argv[2])
        tmp = tj.read_flash_test(addr, DMA_WORD)
        print(hex(tmp))
    elif cmd == "earse":
        addr = str_to_int(sys.argv[2])
        tj.erase_flash(addr)
    elif cmd == "write":
        addr = str_to_int(sys.argv[2])
        value = str_to_int(sys.argv[3])
        tj.unlock_bypass()
        tj.write_flash_bypass_no_earse(addr, value)
    elif cmd == "cfe":
        if len(sys.argv) == 3:
            target = sys.argv[2]
        else:
            target = "%d.bin"%int(time.time())
        start_addr = 0x1FC00000
        length = 0x40000
        data = ""
        for x in range(0, length, 4):
            tmp = tj.read_flash_test(start_addr + x, DMA_WORD)
            data += struct.pack("I", tmp)
        with open("/tmp/%s"%target, "wb") as f:
            f.write(data)
    elif cmd == "flash_cfe":
        length = 0x40000
        block_length = 0x2000
        start_addr = 0x1FC00000
        # for x in range(0, length, block_length):
        #     print("start earse 0x%x"%(start_addr + x))
        #     tj.erase_flash(start_addr + x)
        tj.unlock_bypass()
        with open("./cfe.bin", "rb") as f:
            data = f.read()
        for x in range(0x2fbc8, length, 4):
            tmp = data[x:x+4]
            tmp = struct.unpack("I", tmp)[0]
            tj.write_flash_bypass_no_earse(start_addr + x, tmp)
            check = tj.read_flash_test(start_addr + x, DMA_WORD)
            if check != tmp:
                print(hex(x), "fail")
                raw_input()
            if x % 0x2000 == 0:
                print("0x%x Done!"%x)
    elif cmd == "check":
        length = 0x40000
        start_addr = 0x1FC00000
        with open("./cfe.bin", "rb") as f:
            data = f.read()
        for x in range(0, length, 4):
            check = tj.read_flash_test(start_addr + x, DMA_WORD)
            tmp = struct.unpack("I", data[x:x+4])[0]
            if check != tmp:
                print(hex(x), "fail")
        

if __name__ == "__main__":
    main()